import re
from copy import deepcopy
from typing import Union

import sqlparse
from pypika import JoinType
from pypika.terms import LiteralValue
from sqlparse.tokens import Name, Punctuation

from metrics_layer.core.model.base import MetricsLayerBase
from metrics_layer.core.model.join import Join, ZenlyticJoinType
from metrics_layer.core.sql.query_filter import MetricsLayerFilter


class MetricsLayerQueryBase(MetricsLayerBase):
    def _base_query(self):
        return self.query_lookup[self.query_type]

    def get_where_with_aliases(
        self, filters: list, project, cte_alias_lookup: dict = {}, raise_if_not_in_lookup: bool = False
    ):
        where = []
        for filter_clause in filters:
            filter_clause["query_type"] = self.query_type
            f = MetricsLayerFilter(
                definition=filter_clause, design=None, filter_type="where", project=project
            )
            where.append(
                f.sql_query(
                    alias_query=True,
                    cte_alias_lookup=cte_alias_lookup,
                    raise_if_not_in_lookup=raise_if_not_in_lookup,
                )
            )
        return where

    @staticmethod
    def parse_identifiers_from_clause(clause: str):
        if clause is None:
            return []
        generator = list(sqlparse.parse(clause)[0].flatten())

        field_names = []
        for i, token in enumerate(generator):
            not_already_added = i == 0 or str(generator[i - 1]) != "."
            if token.ttype == Name and not_already_added:
                field_names.append(str(token))

            if token.ttype == Punctuation and str(token) == ".":
                if generator[i - 1].ttype == Name and generator[i + 1].ttype == Name:
                    field_names[-1] += f".{str(generator[i+1])}"
        return field_names

    @staticmethod
    def get_pypika_join_type(join: Join):
        return MetricsLayerQueryBase.pypika_join_type_lookup(join.join_type)

    @staticmethod
    def pypika_join_type_lookup(join_type: str):
        if join_type == ZenlyticJoinType.left_outer:
            return JoinType.left
        elif join_type == ZenlyticJoinType.inner:
            return JoinType.inner
        elif join_type == ZenlyticJoinType.full_outer:
            return JoinType.outer
        elif join_type == ZenlyticJoinType.cross:
            return JoinType.cross
        return JoinType.left

    @staticmethod
    def sql(sql: str, alias: Union[None, str] = None):
        if alias:
            return LiteralValue(sql + f" as {alias}")
        return LiteralValue(sql)

    @staticmethod
    def strip_alias(sql: str):
        stripped_sql = deepcopy(sql)
        matches = re.findall(r"(?i)\ as\ ", stripped_sql)
        if matches:
            alias = " AS "
            for match in matches:
                stripped_sql = stripped_sql.replace(match, alias)
            return alias.join(stripped_sql.split(alias)[:-1])
        return sql


class QueryKindTypes:
    merged = "MERGED"
    single = "SINGLE"



import os

from .project_reader_base import ProjectReaderBase


class MetricsLayerProjectReader(ProjectReaderBase):
    def load(self) -> tuple:
        models, views, dashboards, topics = [], [], [], []

        model_folders = self.get_folders("model-paths")
        view_folders = self.get_folders("view-paths")
        dashboard_folders = self.get_folders("dashboard-paths", raise_errors=False)
        topic_folders = self.get_folders("topic-paths", raise_errors=False)
        all_folders = model_folders + view_folders + dashboard_folders + topic_folders

        file_names = self.search_for_yaml_files(all_folders)

        for fn in file_names:
            yaml_dict = self.read_yaml_file(fn)
            yaml_dict["_file_path"] = os.path.relpath(fn, start=self.repo.folder)

            # Handle keyerror
            if "type" not in yaml_dict and "zenlytic_project" not in fn:
                print(f"WARNING: file {fn} is missing a type")

            yaml_type = yaml_dict.get("type")

            if yaml_type == "model":
                models.append(yaml_dict)
            elif yaml_type == "view":
                views.append(yaml_dict)
            elif yaml_type == "dashboard":
                dashboards.append(yaml_dict)
            elif yaml_type == "topic":
                topics.append(yaml_dict)
            elif yaml_type:
                print(
                    f"WARNING: Unknown file type '{yaml_type}' options are 'model', 'view', 'dashboard', "
                    "or 'topic'"
                )

        return models, views, dashboards, topics

    def search_for_yaml_files(self, folders: list):
        file_names = self.repo.search("*.yml", folders) + self.repo.search("*.yaml", folders)
        return list(set(file_names))

    def get_folders(self, key: str, default: str = None, raise_errors: bool = True):
        if not self.zenlytic_project:
            return []

        if key in self.zenlytic_project:
            return [self._abs_path(p) for p in self.zenlytic_project[key]]
        elif raise_errors:
            raise KeyError(
                f"Missing required key '{key}' in zenlytic_project.yml \n"
                "Learn more about setting these keys here: https://docs.zenlytic.com"
            )
        return []

    def _abs_path(self, path: str):
        if not os.path.isabs(path):
            path = os.path.join(self.repo.folder, path)
        return path



from metrics_layer.core.exceptions import QueryError
from metrics_layer.core.model.definitions import Definitions
from metrics_layer.core.sql.query_cumulative_metric import CumulativeMetricsQuery
from metrics_layer.core.sql.query_design import MetricsLayerDesign
from metrics_layer.core.sql.query_funnel import FunnelQuery
from metrics_layer.core.sql.query_generator import MetricsLayerQuery
from metrics_layer.core.utils import flatten_filters


class SingleSQLQueryResolver:
    def __init__(
        self,
        metrics: list,
        dimensions: list = [],
        funnel: dict = {},
        where: str = None,  # Either a list of json or a string
        having: str = None,  # Either a list of json or a string
        order_by: str = None,  # Either a list of json or a string
        topic=None,
        model=None,
        project=None,
        **kwargs,
    ):
        self.field_lookup = {}
        self.no_group_by = False
        self.has_cumulative_metric = False
        self.verbose = kwargs.get("verbose", False)
        self.select_raw_sql = kwargs.get("select_raw_sql", [])
        self.suppress_warnings = kwargs.get("suppress_warnings", False)
        self.limit = kwargs.get("limit")
        self.return_pypika_query = kwargs.get("return_pypika_query")
        self.force_group_by = kwargs.get("force_group_by", False)
        self.project = project
        self.metrics = metrics
        self.dimensions = dimensions
        self.funnel, self.is_funnel_query = self.parse_funnel(funnel)
        self.topic = topic
        self.model = model
        self.parse_field_names(where, having, order_by)
        self.nesting_depth = kwargs.get("nesting_depth", 0)
        self.query_type = kwargs.get("query_type")
        if self.query_type is None:
            raise QueryError(
                "Could not determine query_type. Please have connection information for "
                "your warehouse in the configuration or explicitly pass the "
                "'query_type' argument to this function"
            )
        self.parse_input()

    def get_query(self, semicolon: bool = True):
        self.design = MetricsLayerDesign(
            no_group_by=self.no_group_by,
            query_type=self.query_type,
            field_lookup=self.field_lookup,
            topic=self.topic,
            model=self.model,
            project=self.project,
        )

        query_definition = {
            "metrics": self.metrics,
            "dimensions": self.dimensions,
            "funnel": self.funnel,
            "where": self.parse_where(self.where),
            "having": self.parse_having(self.having),
            "order_by": self.order_by,
            "select_raw_sql": self.select_raw_sql,
            "limit": self.limit,
            "return_pypika_query": self.return_pypika_query,
            "nesting_depth": self.nesting_depth,
        }
        if self.has_cumulative_metric and self.is_funnel_query:
            raise QueryError("Cumulative metrics cannot be used with funnel queries")

        elif self.has_cumulative_metric:
            query_generator = CumulativeMetricsQuery(
                query_definition, design=self.design, suppress_warnings=self.suppress_warnings
            )

        elif self.is_funnel_query:
            query_generator = FunnelQuery(
                query_definition, design=self.design, suppress_warnings=self.suppress_warnings
            )

        else:
            query_generator = MetricsLayerQuery(
                query_definition, design=self.design, suppress_warnings=self.suppress_warnings
            )

        # If query type does not allow semicolons
        if self.query_type in Definitions.no_semicolon_warehouses:
            semicolon = False
        query = query_generator.get_query(semicolon=semicolon)

        return query

    def get_used_views(self):
        unique_view_names = {f.view.name for f in self.field_lookup.values()}
        return [self.project.get_view(name) for name in unique_view_names]

    def parse_where(self, where: list):
        if where is None or where == [] or self._is_literal(where):
            return where
        where_with_query = []
        for w in where:
            if "funnel" in w:
                non_funnel_where = [f for f in where if "funnel" not in f and "group_by" not in f]
                funnel_query = {
                    "metrics": self.metrics,
                    "dimensions": [],
                    "funnel": w["funnel"],
                    "where": non_funnel_where,
                    "having": self.having,
                }
                w["query_class"] = FunnelQuery(
                    funnel_query, design=self.design, suppress_warnings=self.suppress_warnings
                )

            if "conditions" in w:
                conditions = w["conditions"]
            else:
                conditions = []
            if conditions:
                field_types = set(
                    [
                        self.field_lookup[f["field"]].field_type
                        for f in self.flatten_filters(conditions)
                        if "group_by" not in f
                    ]
                )
                if "measure" in field_types and (
                    "dimension" in field_types or "dimension_group" in field_types
                ):
                    raise QueryError(
                        "Cannot mix dimensions and measures in a compound filter with a logical_operator"
                    )
            where_with_query.append(w)
        return where_with_query

    def parse_having(self, having: list):
        if having is None or having == [] or self._is_literal(having):
            return having
        validated_having = []
        for h in having:
            if "conditions" in h:
                conditions = h["conditions"]
            else:
                conditions = []
            if conditions:
                field_types = set(
                    [self.field_lookup[f["field"]].field_type for f in self.flatten_filters(conditions)]
                )
                if "measure" in field_types and (
                    "dimension" in field_types or "dimension_group" in field_types
                ):
                    raise QueryError(
                        "Cannot mix dimensions and measures in a compound filter with a logical_operator"
                    )
            validated_having.append(h)
        return validated_having

    def parse_input(self):
        all_field_names = self.metrics + self.dimensions
        if len(set(all_field_names)) != len(all_field_names):
            raise QueryError("Ambiguous field names in the metrics and dimensions")

        for name in self.metrics:
            field = self.get_field_with_error_handling(name, "Metric")
            if field.is_cumulative():
                self.has_cumulative_metric = True
            self.field_lookup[name] = field

        metric_view = None if len(self.metrics) == 0 else self.field_lookup[self.metrics[0]].view.name

        for name in self.dimensions:
            field = self.get_field_with_error_handling(name, "Dimension")
            # We will not use a group by if the primary key of the main resulting table is included
            if field.primary_key and field.view.name == metric_view and not self.force_group_by:
                self.no_group_by = True
            self.field_lookup[name] = field

        for name in self._where_field_names:
            self.field_lookup[name] = self.get_field_with_error_handling(name, "Where clause field")

        for name in self._having_field_names:
            self.field_lookup[name] = self.get_field_with_error_handling(name, "Having clause field")

        for name in self._order_by_field_names:
            self.field_lookup[name] = self.get_field_with_error_handling(name, "Order by field")

    def get_field_with_error_handling(self, field_name: str, error_prefix: str):
        field = self.project.get_field(field_name, model=self.model)
        if field is None:
            raise QueryError(f"{error_prefix} {field_name} not found")
        return field

    def parse_field_names(self, where, having, order_by):
        self.where = self._check_for_dict(where)
        if self._is_literal(self.where):
            self._where_field_names = MetricsLayerQuery.parse_identifiers_from_clause(self.where)
        else:
            self._where_field_names = self.parse_identifiers_from_dicts(self.where)

        self.having = self._check_for_dict(having)
        if self._is_literal(self.having):
            self._having_field_names = MetricsLayerQuery.parse_identifiers_from_clause(self.having)
        else:
            self._having_field_names = self.parse_identifiers_from_dicts(self.having)

        self.order_by = self._check_for_dict(order_by)
        if self._is_literal(self.order_by):
            self._order_by_field_names = MetricsLayerQuery.parse_identifiers_from_clause(self.order_by)
        else:
            self._order_by_field_names = self.parse_identifiers_from_dicts(self.order_by)
        return self._where_field_names, self._having_field_names, self._order_by_field_names

    def parse_funnel(self, funnel: dict):
        if funnel != {}:
            if all(k in funnel for k in ["steps", "within"]):
                is_funnel_query = True
            else:
                raise QueryError("Funnel query must have 'steps' and 'within' keys")
        else:
            is_funnel_query = False
        return funnel, is_funnel_query

    @staticmethod
    def _is_literal(clause):
        return isinstance(clause, str) or clause is None

    def parse_identifiers_from_dicts(self, conditions: list):
        flattened_conditions = SingleSQLQueryResolver.flatten_filters(conditions)
        try:
            field_names = []
            for cond in flattened_conditions:
                if "group_by" in cond:
                    field_names.append(cond["group_by"])
                else:
                    field_names.append(cond["field"])
                if "value" in cond and isinstance(cond["value"], str):
                    mapped_field = self.project.get_mapped_field(cond["value"], model=self.model)
                    if mapped_field:
                        field_names.append(cond["value"])
            return field_names
        except KeyError:
            for cond in conditions:
                if "field" not in cond:
                    break
            raise QueryError(f"Identifier was missing required 'field' key: {cond}")

    @staticmethod
    def flatten_filters(filters: list, return_nesting_depth: bool = False):
        return flatten_filters(filters, return_nesting_depth=return_nesting_depth)

    @staticmethod
    def _check_for_dict(conditions: list):
        if isinstance(conditions, dict):
            return [conditions]
        return conditions



import functools
from copy import deepcopy

from pypika import Criterion, JoinType, Table

from metrics_layer.core.exceptions import AccessDeniedOrDoesNotExistException
from metrics_layer.core.model.definitions import Definitions
from metrics_layer.core.model.filter import LiteralValueCriterion
from metrics_layer.core.sql.query_base import MetricsLayerQueryBase
from metrics_layer.core.sql.query_design import MetricsLayerDesign
from metrics_layer.core.sql.query_dialect import query_lookup
from metrics_layer.core.sql.query_filter import MetricsLayerFilter
from metrics_layer.core.sql.query_generator import MetricsLayerQuery

SNOWFLAKE_DATE_SPINE = (
    "select dateadd(day, seq4(), '2000-01-01') as date from table(generator(rowcount => 365*40))"
)
BIGQUERY_DATE_SPINE = "select date from unnest(generate_date_array('2000-01-01', '2040-01-01')) as date"
POSTGRES_DATE_SPINE = (
    "select date from generate_series('2000-01-01'::date, '2040-01-01'::date, '1 day') as date"
)


class CumulativeMetricsQuery(MetricsLayerQueryBase):
    """ """

    def __init__(self, definition: dict, design: MetricsLayerDesign, suppress_warnings: bool = False) -> None:
        self.design = design
        self.query_type = self.design.query_type
        self.no_group_by = self.design.no_group_by
        self.query_lookup = query_lookup
        self.suppress_warnings = suppress_warnings

        self.date_spine_cte_name = design.date_spine_cte_name
        self.base_cte_name = design.base_cte_name

        self._default_date_memo = {}
        super().__init__(definition)

    def __hash__(self):
        return hash(self.design.project)

    def get_query(self, semicolon: bool = True):
        self.cumulative_metrics, self.non_cumulative_metrics = self.separate_metrics()
        has_non_cumulative_metrics = len(self.non_cumulative_metrics) > 0

        base_cte_query = self._base_query()

        date_spine_query = self.date_spine(self.query_type)

        base_cte_query = base_cte_query.with_(Table(date_spine_query), self.date_spine_cte_name)

        for cumulative_metric in self.cumulative_metrics:
            subquery, subquery_alias = self.cumulative_subquery(cumulative_metric)
            base_cte_query = base_cte_query.with_(Table(subquery), subquery_alias)
            aggregate_subquery, aggregate_alias = self.aggregate_cumulative_subquery(cumulative_metric)
            base_cte_query = base_cte_query.with_(Table(aggregate_subquery), aggregate_alias)

        if has_non_cumulative_metrics:
            query, base_cte_name = self.non_cumulative_subquery()
            base_cte_query = base_cte_query.with_(Table(query), base_cte_name)

        if has_non_cumulative_metrics:
            base = Table(self.base_cte_name)
            cumulative_idx = 0
        else:
            base = Table(self.cumulative_metrics[0].cte_prefix())
            cumulative_idx = 1

        base_cte_query = base_cte_query.from_(base)

        for cumulative_metric in self.cumulative_metrics[cumulative_idx:]:
            criteria, cte_alias = self._derive_join(cumulative_metric, base)
            base_cte_query = base_cte_query.join(Table(cte_alias), JoinType.left).on(criteria)

        select = self._get_select_columns(base)
        base_cte_query = base_cte_query.select(*select)

        if self.having:
            where = self.get_where_with_aliases(self.having, project=self.design)
            base_cte_query = base_cte_query.where(Criterion.all(where))

        base_cte_query = base_cte_query.limit(self.limit)
        sql = str(base_cte_query)
        if semicolon:
            sql += ";"
        return sql

    def separate_metrics(self):
        cumulative_metrics, non_cumulative_metrics = [], []
        for field_name in self.metrics + self.having:
            if isinstance(field_name, str):
                field = self.design.get_field(field_name)
            else:
                field = self.design.get_field(field_name["field"])

            if field.is_cumulative():
                if field.type == "number":
                    for reference in field.referenced_fields(field.sql):
                        if reference.is_cumulative():
                            cumulative_metrics.append(reference)
                        else:
                            non_cumulative_metrics.append(reference)
                else:
                    cumulative_metrics.append(field)
            else:
                non_cumulative_metrics.append(field)

        non_cumulative_metrics = self.design.deduplicate_fields(non_cumulative_metrics)
        cumulative_metrics = self.design.deduplicate_fields(cumulative_metrics)
        return cumulative_metrics, non_cumulative_metrics

    @staticmethod
    def date_spine(query_type: str):
        if query_type in {Definitions.snowflake, Definitions.redshift}:
            return SNOWFLAKE_DATE_SPINE
        elif query_type == Definitions.bigquery:
            return BIGQUERY_DATE_SPINE
        elif query_type == Definitions.postgres:
            return POSTGRES_DATE_SPINE
        raise NotImplementedError(f"Database {query_type} not implemented yet")

    def date_spine_by_time_frame(self):
        dimension_group = self.default_date_dimension_group()
        valid = {"date", "week", "month", "quarter", "year"}
        if dimension_group not in valid:
            dimension_group = "date"
            date_name = self.cumulative_metrics[0].measure.view.default_date
            key = f"{self.cumulative_metrics[0].measure.view.name}.{date_name}_{dimension_group}"
            self.dimensions.append(key)
            return self.date_spine_cte_name
        elif dimension_group == "date":
            return self.date_spine_cte_name
        else:
            referenced_metric = self.cumulative_metrics[0].measure
            date_field = self._get_default_date(referenced_metric, self.cumulative_metrics[0])
            query = self._base_query()
            select = self.sql(
                date_field.apply_dimension_group_time_sql("date", query_type=self.query_type), alias="date"
            )
            query = query.from_(Table(self.date_spine_cte_name)).select(select).distinct()
            return query

    def cumulative_subquery(self, cumulative_metric):
        cumulative_metric_cte_alias = cumulative_metric.cte_prefix(aggregated=False)
        reference_metric = cumulative_metric.measure
        dimensions = [self.design.get_field(d) for d in self.dimensions]
        date_field = self._get_default_date(reference_metric, cumulative_metric)
        if not any(date_field.name == d.name and date_field.view.name == d.view.name for d in dimensions):
            dimensions.append(date_field)

        if cumulative_metric.cumulative_where:
            dimensions.extend(self._cumulative_where_fields(cumulative_metric))
        dimension_ids = [d.id() for d in dimensions]

        where = []
        for w in self.where:
            where_field = self.design.get_field(w["field"])
            if not self._is_default_date(where_field):
                where.append(w)

        query = self._subquery(
            metrics=[reference_metric], dimensions=dimension_ids, where=where, no_group_by=True
        )
        return query, cumulative_metric_cte_alias

    def non_cumulative_subquery(self):
        query = self._subquery(
            metrics=self.non_cumulative_metrics,
            dimensions=self.dimensions,
            where=self.where,
            no_group_by=False,
        )
        return query, self.base_cte_name

    def _subquery(self, metrics: list, dimensions: list, where: list, no_group_by: bool):
        sub_definition = deepcopy(self._definition)
        sub_definition["metrics"] = [metric.id() for metric in metrics]
        sub_definition["dimensions"] = dimensions
        sub_definition["where"] = where
        sub_definition["having"] = []
        sub_definition["limit"] = None

        field_lookup = {k: v for k, v in self.design.field_lookup.items()}
        self.design.field_lookup = {k: v for k, v in field_lookup.items() if v.field_type != "measure"}
        self.design.no_group_by = no_group_by
        for metric in metrics:
            self.design.field_lookup[metric.id()] = metric
        self.design.joins.cache_clear()
        self.design.functional_pk.cache_clear()

        query_generator = MetricsLayerQuery(
            sub_definition, design=self.design, suppress_warnings=self.suppress_warnings
        )
        query = query_generator.get_query(semicolon=False)
        self.design.no_group_by = False
        self.design.field_lookup = field_lookup
        self.design.joins.cache_clear()
        self.design.functional_pk.cache_clear()

        return query

    def aggregate_cumulative_subquery(self, cumulative_metric):
        cte_alias = cumulative_metric.cte_prefix(aggregated=False)

        referenced_metric = cumulative_metric.measure
        date_field = self._get_default_date(referenced_metric, cumulative_metric)

        from_query = self._base_query()
        date_spine = self.date_spine_by_time_frame()
        if date_spine != self.date_spine_cte_name:
            date_table = Table(f"({date_spine})", alias=self.date_spine_cte_name)
        else:
            date_table = Table(self.date_spine_cte_name)
        from_query = from_query.from_(date_table)

        date_spine_reference = f"{self.date_spine_cte_name}.date"
        less_than_now = f"{cte_alias}.{date_field.alias(with_view=True)}<={date_spine_reference}"

        # For a cumulative filter like setting cancelled date to less than the date_spine value
        if cumulative_metric.cumulative_where:
            c_where = self._replace_cumulative_where(cumulative_metric, date_spine_reference)
            less_than_now = less_than_now + " AND " + c_where

        # E.g. for a 7 day window
        # window = f"{cte_alias}.{date_field_name}>DATEADD(day, -7, {self.date_spine_cte_name}.date)"
        criteria = LiteralValueCriterion(less_than_now)
        from_query = from_query.join(Table(cte_alias), JoinType.inner).on(criteria)

        select = []
        default_date_is_present, group_by = False, []
        for field_name in [referenced_metric.id()] + self.dimensions:
            field = self.design.get_field(field_name)
            # Date field is the default date but not the default date on the referenced metric
            # I need to check if the date field is the default date on any other metric in the query,
            # and if so use the date spine instead of the date field itself
            if self._is_default_date(field):
                field_sql = date_spine_reference
                default_date_is_present = True
            else:
                field_sql = field.sql_query(query_type=self.query_type, alias_only=True)
                if field.field_type != "measure":
                    group_by.append(self.sql(field_sql))
            select.append(self.sql(field_sql, alias=field.alias(with_view=True)))

        from_query = from_query.select(*select)

        from_query = from_query.where(LiteralValueCriterion(f"{date_spine_reference}<={self.current_date}"))

        having = []
        for w in self.where:
            where_field = self.design.get_field(w["field"])
            if self._is_default_date(where_field):
                date_having = deepcopy(w)
                dimension_group = deepcopy(date_field.dimension_group)
                date_field.dimension_group = where_field.dimension_group
                date_having["field"] = date_field.id()
                date_having["query_type"] = self.query_type
                f = MetricsLayerFilter(definition=date_having, design=self.design, filter_type="having")
                date_spine_sql = date_field.apply_dimension_group_time_sql(
                    date_spine_reference, self.query_type
                )
                date_field.dimension_group = dimension_group
                having.append(f.criterion(date_spine_sql))

        if default_date_is_present:
            group_by.append(self.sql(date_spine_reference))

        if group_by:
            from_query = from_query.groupby(*group_by)

        if having and default_date_is_present:
            from_query = from_query.having(LiteralValueCriterion(Criterion.all(having)))

        return from_query, cumulative_metric.cte_prefix()

    def _is_default_date(self, field):
        date_aliases = []
        for cumulative_metric in self.cumulative_metrics:
            date_name = cumulative_metric.measure.view.default_date
            if "." in date_name:
                date_aliases.append(date_name)
            else:
                date_aliases.append(f"{cumulative_metric.measure.view.name}.{date_name}")
        return f"{field.view.name}.{field.name}" in date_aliases

    @functools.lru_cache(maxsize=None)
    def default_date_dimension_group(self):
        for dimension_name in self.dimensions:
            dimension = self.design.get_field(dimension_name)
            if self._is_default_date(dimension):
                return dimension.dimension_group
        return "date"

    def _replace_cumulative_where(self, cumulative_metric, date_spine_reference: str):
        cte_prefix = cumulative_metric.cte_prefix(aggregated=False)
        dimension_group = self.default_date_dimension_group()
        c_where = cumulative_metric.cumulative_where
        replaced_where = c_where.replace("${cumulative_date}", date_spine_reference)

        for ref in self._cumulative_where_fields(cumulative_metric, refs_only=True):
            field = self.design.get_field(ref)

            if (
                field.field_type == "dimension_group"
                and field.dimension_group != dimension_group
                and cumulative_metric.update_where_timeframe
            ):
                field.dimension_group = dimension_group

            sql = field.sql_query(query_type=self.query_type, alias_only=True)
            replaced_where = replaced_where.replace("${" + ref + "}", f"{cte_prefix}.{sql}")

        return replaced_where

    def _cumulative_where_fields(self, cumulative_metric, refs_only=False):
        dimension_group = self.default_date_dimension_group()
        c_where = deepcopy(cumulative_metric.cumulative_where)
        c_where = c_where.replace("${cumulative_date}", "")
        refs = cumulative_metric.fields_to_replace(c_where)
        if refs_only:
            return refs

        fields = []
        for r in refs:
            field = self.design.get_field(r)
            if (
                field.field_type == "dimension_group"
                and field.dimension_group != dimension_group
                and cumulative_metric.update_where_timeframe
            ):
                field = self.design.get_field(f"{field.view.name}.{field.name}_{dimension_group}")
            fields.append(field)
        return fields

    def _get_default_date(self, field, cumulative_metric):
        date_name = field.view.default_date
        dimension_group = self.default_date_dimension_group()
        date_field_name = f"{date_name}_{dimension_group}"
        date_key = f"{field.view.name}.{date_field_name}"
        if date_key in self._default_date_memo:
            return self._default_date_memo[date_key]
        try:
            self._default_date_memo[date_key] = self.design.get_field(date_key)
        except Exception:
            raise AccessDeniedOrDoesNotExistException(
                (
                    f"Could not find date needed to aggregate cumulative metric {cumulative_metric.name}, "
                    f"looking for date field named {date_field_name} in view {field.view.name}"
                ),
                object_type="field",
                object_name=date_field_name,
            )
        return self._default_date_memo[date_key]

    def _derive_join(self, cumulative_metric, base_table: Table):
        cte_alias = cumulative_metric.cte_prefix()

        conditions = []
        for field_name in self.dimensions:
            field = self.design.get_field(field_name)
            field_alias = field.alias(with_view=True)
            conditions.append(f"{base_table.get_table_name()}.{field_alias}={cte_alias}.{field_alias}")

        if len(conditions) > 0:
            raw_criteria = " and ".join(conditions)
        else:
            raw_criteria = "1=1"

        criteria = LiteralValueCriterion(raw_criteria)
        return criteria, cte_alias

    def _get_select_columns(self, base_table: Table):
        select = []
        for field_name in self.dimensions + self.metrics:
            field = self.design.get_field(field_name)
            if field.type == "number" and field.is_cumulative():
                field_sql = field.sql_query(query_type=self.query_type, alias_only=True)
            elif field.is_cumulative():
                field_sql = field.measure.alias(with_view=True)
            else:
                field_sql = field.alias(with_view=True)

            if not field.is_cumulative():
                field_sql = f"{base_table.get_table_name()}.{field_sql}"
            elif field.type != "number":
                field_sql = f"{field.cte_prefix()}.{field_sql}"

            select.append(MetricsLayerQuery.sql(field_sql, alias=field.alias(with_view=True)))
        return select

    @property
    def current_date(self):
        if self.query_type == Definitions.snowflake:
            return "current_date()"
        elif self.query_type == Definitions.redshift:
            return "current_date()"
        elif self.query_type == Definitions.bigquery:
            return "current_date()"
        elif self.query_type == Definitions.postgres:
            return "current_date"
        else:
            raise NotImplementedError(f"Query type {self.query_type} not supported")



from collections import Counter, defaultdict
from copy import deepcopy
from typing import List, Union

from metrics_layer.core.exceptions import JoinError, QueryError
from metrics_layer.core.model.filter import Filter, MetricsLayerFilterExpressionType
from metrics_layer.core.model.project import Project
from metrics_layer.core.sql.merged_query_resolve import MergedSQLQueryResolver
from metrics_layer.core.sql.query_base import QueryKindTypes
from metrics_layer.core.sql.single_query_resolve import SingleSQLQueryResolver


class SQLQueryResolver(SingleSQLQueryResolver):
    def __init__(
        self,
        metrics: list,
        dimensions: list = [],
        funnel: dict = {},  # A dict with steps (list) and within (dict)
        where: Union[str, None, List] = None,  # Either a list of json or a string
        having: Union[str, None, List] = None,  # Either a list of json or a string
        order_by: Union[str, None, List] = None,  # Either a list of json or a string
        project: Union[Project, None] = None,
        connections: List = [],
        **kwargs,
    ):
        self.field_lookup = {}
        self.no_group_by = False
        self.mapping_forces_merged_result = False
        self.verbose = kwargs.get("verbose", False)
        self.select_raw_sql = kwargs.get("select_raw_sql", [])
        self.suppress_warnings = kwargs.get("suppress_warnings", False)
        self.limit = kwargs.get("limit")
        self.single_query = kwargs.get("single_query", False)
        self.kwargs = kwargs
        self.project = project
        if self.kwargs.get("topic"):
            self.topic = self.project.get_topic(kwargs["topic"])
            self.kwargs["topic"] = self.topic
        else:
            self.topic = None
        self.model = self._get_model_for_query(kwargs.get("model_name"), metrics, dimensions)
        self.connections = connections
        self.metrics = metrics
        self.dimensions = dimensions
        self.funnel = funnel
        self.where = where if where else []
        always_where = self._apply_always_filter(metrics + dimensions)
        if always_where:
            self.where.extend(always_where)
        self.where = self._clean_conditional_filter_syntax(self.where)
        self.having = self._clean_conditional_filter_syntax(having)
        self.order_by = order_by
        self.connection = self._get_connection(self.model.connection)
        self.kwargs["query_type"] = self._get_query_type(self.connection, self.kwargs)
        connection_schema = self._get_connection_schema(self.connection)
        self.project.set_connection_schema(connection_schema)
        self.field_id_mapping = {}
        self._resolve_mapped_fields()
        self.query_type = None

    @property
    def is_merged_result(self):
        has_explicit_merge = any(self.project.get_field(m).is_merged_result for m in self.metrics)
        has_specified_merge = self.kwargs.get("merged_result", False)
        return has_explicit_merge or has_specified_merge or self.mapping_forces_merged_result

    def get_query(self, semicolon: bool = True):
        err_msg = ""
        is_explicit_merge = self.is_merged_result
        if not is_explicit_merge:
            try:
                self.query_kind = QueryKindTypes.single
                return self._get_single_query(semicolon=semicolon)
            except JoinError as e:
                err_msg = "Could not execute the query as a single query. Trying as a merged result query."
                if self.single_query:
                    raise e
                if self.verbose:
                    print(err_msg)
        self.query_kind = QueryKindTypes.merged
        try:
            return self._get_merged_result_query(semicolon=semicolon)
        except QueryError as e:
            appended = (
                "Zenlytic tries to merge query results by default if there is no join path between "
                "the views. If you'd like to disable this behavior pass single_query=True to the "
                "function call.\n\nIf you're seeing this and you expected the views to join on a "
                "primary or foreign key, make sure you have the right identifiers set on the views."
            )
            error_message = (
                f"{err_msg}\n\n{appended}\n\n{deepcopy(e.message)}"
                if self.verbose
                else f"{deepcopy(e.message)}"
            )
            e.message = error_message
            raise e

    def _get_single_query(self, semicolon: bool):
        resolver = SingleSQLQueryResolver(
            metrics=self.metrics,
            dimensions=self.dimensions,
            funnel=self.funnel,
            where=self.where,
            having=self.having,
            order_by=self.order_by,
            model=self.model,
            project=self.project,
            **self.kwargs,
        )
        query = resolver.get_query(semicolon)
        self.query_type = resolver.query_type
        return query

    def _get_merged_result_query(self, semicolon: bool):
        resolver = MergedSQLQueryResolver(
            metrics=self.metrics,
            dimensions=self.dimensions,
            funnel=self.funnel,
            where=self.where,
            having=self.having,
            order_by=self.order_by,
            model=self.model,
            project=self.project,
            **self.kwargs,
        )
        query = resolver.get_query(semicolon)
        self.query_type = resolver.query_type
        return query

    def _resolve_mapped_fields(self):
        self.mapping_lookup, self.field_lookup, self.field_object_lookup = {}, {}, {}
        self._where_fields, self._having_fields, self._order_fields = self.parse_field_names(
            self.where, self.having, self.order_by
        )
        self._all_fields = (
            self.metrics + self.dimensions + self._where_fields + self._having_fields + self._order_fields
        )

        replace_bools = [True] * len(self._all_fields) + [False] * len(
            self.kwargs.get("mapping_lookup_dimensions", [])
        )
        lookup_only_fields = [
            d for d in self.kwargs.get("mapping_lookup_dimensions", []) if d not in self._all_fields
        ]
        field_ids = self._all_fields + lookup_only_fields
        for do_replace, field_name in zip(replace_bools, field_ids):
            mapped_field = self.project.get_mapped_field(field_name, model=self.model)
            if mapped_field:
                self.mapping_lookup[field_name] = {**mapped_field, "do_replace": do_replace}
            else:
                field_obj = self.project.get_field(field_name, model=self.model)
                self.field_object_lookup[field_name] = field_obj
                self.field_lookup[field_name] = field_obj.join_graphs()

        if not self.mapping_lookup:
            # We need to call this here because the 'field' value in the
            # group by filter does not show up in the field lookup
            self._replace_field_value_in_group_by_filter()
            return

        if self.field_lookup:
            mergeable_graphs, joinable_graphs = self._join_graphs_by_type(self.field_lookup)
            self._handle_invalid_merged_result(mergeable_graphs, joinable_graphs)
            for name, mapped_field in self.mapping_lookup.items():
                if not mapped_field.pop("do_replace"):
                    continue  # Skip fields that are only for lookup

                is_date_mapping = mapped_field["name"] in self.model.special_mapping_values
                if is_date_mapping and len(self.metrics) >= 1:
                    # Use the first metric to replace the mapped field everywhere
                    # This is fine to do in the case we have multiple metrics because in that
                    # case we'll need to run a merged result query which will take care of mapping the
                    # canon_date's for each metric regardless of the canon_date chosen for the filters
                    first_metric_field = self._get_field_from_lookup(self.metrics[0])
                    canon_date_id = f'{first_metric_field.canon_date}_{mapped_field["name"]}'
                    replace_with = self.project.get_field(canon_date_id)

                    # If the additional metrics have different canon_dates, we'll need to
                    # change the query type to be a merged result
                    for field_name in self.metrics[1:]:
                        metric_field = self._get_field_from_lookup(field_name)
                        if metric_field.canon_date != first_metric_field.canon_date:
                            self.mapping_forces_merged_result = True
                            break
                elif is_date_mapping and len(self.dimensions) >= 1:
                    canon_dates = set()
                    for d in self.dimensions:
                        field = self._get_field_from_lookup(d, only_search_lookup=True)
                        if field and field.canon_date:
                            canon_dates.add(field.canon_date)

                    if len(canon_dates) >= 1:
                        # When there is no measure present, we will chose the canon date
                        # from the dimension that has the most measures
                        if len(canon_dates) > 1:
                            views_by_n_measures = {}
                            for canon_date in canon_dates:
                                view_name = canon_date.split(".")[0]
                                measures = [
                                    f
                                    for f in self.project.get_view(view_name).fields()
                                    if f.field_type == "measure"
                                ]
                                views_by_n_measures[canon_date] = len(measures)
                            sorted_canon_dates = list(
                                sorted(canon_dates, key=lambda x: views_by_n_measures[x], reverse=True)
                            )
                        else:
                            # Sort is not needed if there is only one canon_date
                            sorted_canon_dates = list(canon_dates)

                        canon_date_id = f'{sorted_canon_dates[0]}_{mapped_field["name"]}'
                        replace_with = self.project.get_field(canon_date_id)
                    else:
                        replace_with = self.determine_field_to_replace_with(
                            mapped_field, joinable_graphs, mergeable_graphs
                        )
                else:
                    replace_with = self.determine_field_to_replace_with(
                        mapped_field, joinable_graphs, mergeable_graphs
                    )
                self._replace_mapped_field(name, replace_with)
        else:
            # This is the scenario where we only have mappings and no other fields in the query

            # First, we need to get all join graphs for all fields involved in the mappings.
            # Since we have no "real" fields, we have no basis for required join graphs we
            # must match, yet. This gives us the raw material to derive which join graphs
            # overlap in the mappings.
            check = defaultdict(list)
            for name, mapped_field in self.mapping_lookup.items():
                for field_id in mapped_field["fields"]:
                    ref_field = self.project.get_field(field_id)
                    check[name].append((field_id, ref_field.join_graphs()))

            # Next, we iterate over all mappings and check if there is a valid join path
            # between all fields in the mappings.
            validity = defaultdict(dict)
            # This is done my comparing a single mapping (name), then...
            for name, field_info in check.items():
                # Looking at all fields that are present in that mapping name, and
                # comparing them to all other fields in all other mappings
                for field_id, join_graphs in field_info:
                    passed, points = 0, 0
                    for other_name, other_field_info in check.items():
                        if name != other_name:
                            for other_field_id, other_join_graphs in other_field_info:
                                if field_id != other_field_id:
                                    # If the fields are joinable or mergeable, then we can form a
                                    # valid query, and we can count them as 'passed'
                                    if set(join_graphs).intersection(set(other_join_graphs)):
                                        passed += 1
                                        # It's not enough just to pass though. There are some things we prefer,
                                        # when there is no solid direction on which fields to choose.

                                        # 1. We prefer when fields are in the same view
                                        points += (
                                            1 if field_id.split(".")[0] == other_field_id.split(".")[0] else 0
                                        )

                                        # 2. We prefer when fields can be joined over when they can only be merged
                                        joinable_first = [j for j in join_graphs if "merged_result" not in j]
                                        joinable_second = [
                                            j for j in other_join_graphs if "merged_result" not in j
                                        ]
                                        if set(joinable_first).intersection(set(joinable_second)):
                                            points += 1
                                        # Once we have determined a field pair is valid,
                                        # we can stop checking other fields under that mapping name
                                        break

                    # If at least one field in each of the other mapping names has 'passed' the check,
                    # then we can consider this mapping name as valid, and assign it the points
                    if passed == len(check) - 1:
                        validity[name][field_id] = points

            # If this exists and we have more than one mapping present
            if validity and len(self.mapping_lookup) > 1:
                replaced_mapping = None
                all_items = [
                    (name, *item) for name, field_info in validity.items() for item in field_info.items()
                ]
                # Sort the mappings by points, and then by field_id (so it is consistent in
                # the event fields have the same points), and replace the highest ranking one
                # and set its join graphs as the active ones all other mappings must adhere to.
                for name, field_id, points in sorted(all_items, key=lambda x: (x[-1], x[1]), reverse=True):
                    replaced_mapping = name
                    replace_with = self.project.get_field(field_id)
                    self.field_lookup[name] = replace_with.join_graphs()
                    break

                self._replace_mapped_field(name, replace_with)
            # In the event there is only one mapping, just pick the first field in the options
            elif len(self.mapping_lookup) == 1:
                for name, mapped_field in self.mapping_lookup.items():
                    replaced_mapping = name
                    replace_with = self.project.get_field(mapped_field["fields"][0])
                    self.field_lookup[name] = replace_with.join_graphs()
                self._replace_mapped_field(name, replace_with)
            else:
                raise QueryError("No valid join path found for mapped fields")

            for name, mapped_field in self.mapping_lookup.items():
                if name != replaced_mapping:
                    mergeable_graphs, joinable_graphs = self._join_graphs_by_type(self.field_lookup)
                    self._handle_invalid_merged_result(mergeable_graphs, joinable_graphs)
                    replace_with = self.determine_field_to_replace_with(
                        mapped_field, joinable_graphs, mergeable_graphs
                    )
                    self._replace_mapped_field(name, replace_with)

        # We also have to swap out the field in a potential group by where clause
        self._replace_field_value_in_group_by_filter()

    def _replace_field_value_in_group_by_filter(self):
        if isinstance(self.where, list) and self.field_lookup:
            optimal_join_graph_connection = set.intersection(*map(set, self.field_lookup.values()))
            optimal_join_graph_connection = [
                o for o in optimal_join_graph_connection if "merged_result" not in o
            ]
            flattened_conditions = SingleSQLQueryResolver.flatten_filters(
                self.where, return_nesting_depth=True
            )
            for cond in flattened_conditions:
                if "group_by" in cond:
                    # Only the group by field needs to be joinable or merge-able to the query
                    # The field in the group by where clause is not required to be joinable
                    # to the whole query, just to the group by field
                    group_by_field = self.project.get_field(cond["group_by"], model=self.model)
                    join_graphs = group_by_field.join_graphs()

                    # Here we need to check if the field is a mapped field
                    # If it is, we need to add the underlying field
                    mapped_field = self.project.get_mapped_field(cond["field"], model=self.model)
                    if mapped_field:
                        replace_with = self.determine_field_to_replace_with(
                            mapped_field, optimal_join_graph_connection, join_graphs
                        )
                        self.field_id_mapping[cond["field"]] = replace_with.id()
                        cond["field"] = replace_with.id()
                elif cond["expression"] in {
                    MetricsLayerFilterExpressionType.IsInQuery.value,
                    MetricsLayerFilterExpressionType.IsNotInQuery.value,
                }:
                    defaults = {
                        "project": self.project,
                        "connections": self.connections,
                        "model_name": self.model.name,
                        "return_pypika_query": False,
                    }
                    if "query_type" in self.kwargs:
                        defaults["query_type"] = self.kwargs["query_type"]

                    # This handles the case where the passed filter is incomplete, and
                    # does not apply the filter
                    if "query" not in cond["value"]:
                        continue

                    if "query" in cond["value"] and not isinstance(cond["value"]["query"], dict):
                        raise QueryError(
                            "Subquery filter value for the key 'query' must be a dictionary. It was"
                            f" {cond['value']['query']}"
                        )

                    if "apply_limit" in cond["value"] and not bool(cond["value"]["apply_limit"]):
                        cond["value"]["query"]["limit"] = None

                    if "nesting_depth" in cond and cond["nesting_depth"] > 0:
                        defaults["nesting_depth"] = cond["nesting_depth"]

                    resolver = SQLQueryResolver(**cond["value"]["query"], **defaults)
                    jg_connection = set.intersection(*map(set, resolver.field_lookup.values()))
                    optimal_jg_connection = [o for o in jg_connection if "merged_result" not in o]

                    mapped_field = self.project.get_mapped_field(cond["value"]["field"], model=self.model)
                    if mapped_field:
                        field = self.determine_field_to_replace_with(
                            mapped_field, optimal_jg_connection, jg_connection
                        )
                        self.field_id_mapping[cond["value"]["field"]] = field.id()
                        cond["value"]["field"] = field.id()
                    else:
                        field = self.project.get_field(cond["value"]["field"])
                    if field.id() not in {self.project.get_field(d).id() for d in resolver.dimensions}:
                        raise QueryError(
                            f"Field {field.id()} not found in subquery dimensions {resolver.dimensions}. You"
                            " must specify a dimension that is present in the subquery."
                        )
                    cond["value"]["sql_query"] = resolver.get_query(semicolon=False)

    def _get_field_from_lookup(self, field_name: str, only_search_lookup: bool = False):
        if field_name in self.field_object_lookup:
            metric_field = self.field_object_lookup[field_name]
        elif only_search_lookup:
            return None
        else:
            metric_field = self.project.get_field(field_name, model=self.model)
        return metric_field

    def determine_field_to_replace_with(self, mapped_field, joinable_graphs, mergeable_graphs):
        joinable, mergeable = [], []
        for field_name in mapped_field["fields"]:
            field = self.project.get_field(field_name)
            join_graphs = field.join_graphs()
            if any(g in joinable_graphs for g in join_graphs):
                joinable.append(field)
            elif any(g in mergeable_graphs for g in join_graphs):
                mergeable.append(field)

        if joinable:
            return joinable[0]
        elif mergeable:
            return mergeable[0]
        else:
            raise QueryError(f'No valid join path found for mapped field "{mapped_field["name"]}"')

    def _join_graphs_by_type(self, field_lookup: dict):
        usable_merged_graphs = set.intersection(*map(set, field_lookup.values()))
        usable_joinable_graphs = [s for s in list(usable_merged_graphs) if "merged_result" not in s]
        return usable_merged_graphs, usable_joinable_graphs

    def _handle_invalid_merged_result(self, mergeable_graphs, joinable_graphs):
        # If both of these are empty there is no join overlap and the query cannot be run
        if len(mergeable_graphs) == 0 and len(joinable_graphs) == 0:
            # first we try to find the issue by removing one key from the lookup and seeing
            # if it works with the rest. This produces the most sensible error message.
            for key in self.field_lookup.keys():
                with_one_removed = {k: v for k, v in self.field_lookup.items() if k != key}
                merged, joined = self._join_graphs_by_type(with_one_removed)
                if len(merged) != 0 or len(joined) != 0:
                    error_message = (
                        f"The field {key} could not be either joined into the query "
                        "or mapped and merged into the query as a merged result. \n\n"
                        "Check that you specify joins to join it in, or specify a mapping "
                        "for a query with two tables that cannot be merged"
                        if self.verbose
                        else (
                            f"Error: The field {key} could not be joined or mapped and merged into the query."
                            " Please try to reformat the query. If the error persists, consult the user for"
                            " further guidance."
                        )
                    )
                    raise QueryError(error_message)
            # Otherwise, we have to show this, worse, error message
            all_fields = list(self.field_lookup.keys()) + list(self.mapping_lookup.keys())
            error_message = (
                "The query could not be either joined or mapped and merged into a valid query"
                f" with the fields:\n\n{', '.join(all_fields)}\n\n"
                "Check that those fields can be joined together or are mapped so they can "
                "be merged across tables"
                if self.verbose
                else (
                    "Error: The query could not be either joined or mapped and merged into a valid query"
                    f" with the fields:\n\n{', '.join(all_fields)}\n\n Please try to reformat the query. If"
                    " the error persists, consult the user for further guidance."
                )
            )
            raise QueryError(error_message)

    def _replace_mapped_field(self, to_replace: str, field):
        self.field_id_mapping[to_replace] = field.id()
        if to_replace in self.metrics:
            idx = self.metrics.index(to_replace)
            self.metrics[idx] = field.id()
        if to_replace in self.dimensions:
            idx = self.dimensions.index(to_replace)
            self.dimensions[idx] = field.id()
        if to_replace in self._where_fields:
            self.where = self._replace_dict_or_literal(self.where, to_replace, field)
        if to_replace in self._having_fields:
            self.having = self._replace_dict_or_literal(self.having, to_replace, field)
        if to_replace in self._order_fields:
            self.order_by = self._replace_dict_or_literal(self.order_by, to_replace, field)

        if to_replace not in self._all_fields:
            raise QueryError(f"Could not find mapped field {to_replace} in query")

    def _replace_dict_or_literal(self, where, to_replace, field):
        if self._is_literal(where):
            return where.replace(to_replace, field.id())
        else:
            result = []
            for w in where:
                if "group_by" in w and w["group_by"] == to_replace:
                    result.append({**w, "group_by": field.id()})
                elif "field" in w and w["field"] == to_replace and "value" in w and w["value"] == to_replace:
                    result.append({**w, "field": field.id(), "value": field.id()})
                elif "value" in w and w["value"] == to_replace:
                    result.append({**w, "value": field.id()})
                elif "field" in w and w["field"] == to_replace:
                    result.append({**w, "field": field.id()})
                elif "field" not in w and "conditions" in w:
                    result.append(
                        {**w, "conditions": self._replace_dict_or_literal(w["conditions"], to_replace, field)}
                    )
                else:
                    result.append(w)
            return result

    def _get_model_for_query(self, model_name: str = None, metrics: list = [], dimensions: list = []):
        models = self.project.models()
        # If you specify the model that's top priority
        if model_name:
            return self.project.get_model(model_name)
        # Otherwise, if there's only one option, we use that
        elif len(models) == 1:
            return models[0]
        # Raise an error if the user doesn't have any models defined yet
        elif len(models) == 0:
            raise QueryError(
                "No models found in this data model. Please specify a model "
                "to connect a data warehouse to your data model."
            )
        # Finally, check views for models
        else:
            return self._derive_model(metrics, dimensions)

    def _derive_model(self, metrics: list, dimensions: list):
        all_model_names, mapping_model_names = [], []
        models = self.project.models()
        for f in metrics + dimensions:
            try:
                model_name = self.project.get_field(f).view.model_name
                all_model_names.append(model_name)
            except Exception:
                for model in models:
                    try:
                        self.project.get_mapped_field(f, model=model)
                        mapping_model_names.append(model.name)
                    except Exception:
                        pass
        all_model_names = list(set(all_model_names))
        if not all_model_names:
            if mapping_model_names:
                # In this case, the only fields we recognize are mappings. These can
                # exist in multiple models, so we don't know which to use. Let's make a
                # guess and choose the model containing the most mappings.
                model_counts = Counter(mapping_model_names)
                sorted_models = [m for m, _ in model_counts.most_common()]
                return self.project.get_model(sorted_models[0])
            # Alternatively, we don't recognize any fields. Let's arbitrarily choose the
            # first model.
            return models[0]
        if len(all_model_names) == 1 and (
            len(mapping_model_names) == 0
            or (len(mapping_model_names) > 0 and all_model_names[0] in mapping_model_names)
        ):
            return self.project.get_model(list(all_model_names)[0])
        raise QueryError(
            "More than one model found in this query. Please specify a model to use by either passing the"
            " name of the model using 'model_name' parameter or by setting the `model_name` property on the"
            " view."
        )

    def _get_query_type(self, connection, kwargs: dict):
        if "query_type" in kwargs:
            return kwargs.pop("query_type")
        elif connection:
            return connection.type
        else:
            raise QueryError(
                "Could not determine query_type. Please have connection information for "
                "your warehouse in the configuration or explicitly pass the "
                "'query_type' argument to this function"
            )

    def _apply_always_filter(self, fields: list):
        always_where = []
        to_add = {"week_start_day": self.model.week_start_day, "timezone": self.project.timezone}
        for field_str in fields:
            try:
                field = self.project.get_field(field_str)
                if field.view.always_filter:
                    parsed_filters = []
                    for f in field.view.always_filter:
                        if "." not in f["field"]:
                            f["field"] = f"{field.view.name}.{f['field']}"
                        parsed_filters.extend(Filter({**f, **to_add}).filter_dict(json_safe=True))
                    always_where.extend(parsed_filters)
            # Handle mappings exception
            except Exception:
                pass

        # Apply always filters from the topic (if present)
        if self.topic and self.topic.always_filter:
            for f in self.topic.always_filter:
                if "." not in f["field"]:
                    raise QueryError(
                        f"Always filter field {f['field']} in the topic {self.topic.label} must be a fully"
                        " qualified field name in the format view_name.field_name"
                    )
                always_where.extend(Filter({**f, **to_add}).filter_dict(json_safe=True))

        return self._deduplicate_always_where_filters(always_where)

    @staticmethod
    def _deduplicate_always_where_filters(filters: list):
        seen = set()
        cleaned_filters = []
        for f in filters:
            hashable_filter = tuple((k, v if not isinstance(v, list) else tuple(v)) for k, v in f.items())
            if hashable_filter not in seen:
                seen.add(hashable_filter)
                cleaned_filters.append(f)
        return cleaned_filters

    def _clean_conditional_filter_syntax(self, filters: Union[str, None, List]):
        if not filters or isinstance(filters, str):
            return filters

        if isinstance(filters, dict):
            return [filters]

        def process_filter(filter_obj):
            if isinstance(filter_obj, dict):
                if "conditional_filter_logic" in filter_obj:
                    return filter_obj["conditional_filter_logic"]
                elif "conditions" in filter_obj:
                    filter_obj["conditions"] = [process_filter(cond) for cond in filter_obj["conditions"]]
            return filter_obj

        return [process_filter(filter_obj) for filter_obj in filters]

    def _get_connection_schema(self, connection):
        if connection is not None:
            return getattr(connection, "schema", None)
        return None

    def _get_connection(self, connection_name: str):
        return next((c for c in self.connections if c.name == connection_name), None)



import json
import os
from copy import deepcopy
from typing import Any

from metrics_layer.core.model.definitions import Definitions
from metrics_layer.core.sql.query_errors import ArgumentError


class MetricsLayerConnectionError(Exception):
    pass


class ConnectionType:
    snowflake = Definitions.snowflake
    bigquery = Definitions.bigquery
    redshift = Definitions.redshift
    postgres = Definitions.postgres
    mysql = Definitions.mysql
    druid = Definitions.druid
    sql_server = Definitions.sql_server
    duck_db = Definitions.duck_db
    databricks = Definitions.databricks
    azure_synapse = Definitions.azure_synapse
    trino = Definitions.trino


class BaseConnection:
    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}>"


class SnowflakeConnection(BaseConnection):
    def __init__(
        self,
        name: str,
        account: str,
        password: str,
        role: str = None,
        username: str = None,
        user: str = None,
        warehouse: str = None,
        database: str = None,
        schema: str = None,
        **kwargs,
    ) -> None:
        self.type = ConnectionType.snowflake
        self.name = name
        self.account = account
        if user and username:
            raise ArgumentError(
                "Received arguments for both user and username, "
                "please send only one argument for the Snowflake user"
            )
        elif username:
            self.username = username
        elif user:
            self.username = user
        elif kwargs.get("auth_method") != "SSO":
            raise ArgumentError("Received no argument for the Snowflake user, pass either user or username")
        else:
            self.username = username
        self.password = password
        self.role = role
        self.warehouse = warehouse
        self.database = database
        self.schema = schema

    def to_dict(self):
        """Dict for use with the snowflake connector"""
        base = {
            "name": self.name,
            "user": self.username,
            "password": self.password,
            "account": self.account,
            "type": self.type,
        }
        if self.warehouse:
            base["warehouse"] = self.warehouse
        if self.database:
            base["database"] = self.database
        if self.schema:
            base["schema"] = self.schema
        if self.role:
            base["role"] = self.role
        return base

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("password")
        attributes["name"] = self.name
        sort_order = ["name", "type", "account", "user", "database", "schema", "warehouse", "role"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}


class DatabricksConnection(BaseConnection):
    def __init__(
        self,
        name: str,
        host: str,
        http_path: str,
        username: str = None,
        password: str = None,
        personal_access_token: str = None,
        client_id: str = None,
        client_secret: str = None,
        database: str = None,
        schema: str = None,
        **kwargs,
    ) -> None:
        self.type = ConnectionType.databricks
        self.name = name
        self.host = host
        self.http_path = http_path
        self.username = username
        self.password = password
        self.personal_access_token = personal_access_token
        self.client_id = client_id
        self.client_secret = client_secret
        self.database = database
        self.schema = schema

    def to_dict(self):
        base = {
            "name": self.name,
            "host": self.host,
            "http_path": self.http_path,
            "username": self.username,
            "password": self.password,
            "personal_access_token": self.personal_access_token,
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "type": self.type,
        }
        if self.database:
            base["database"] = self.database
        if self.schema:
            base["schema"] = self.schema
        return base

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("password")
        attributes["name"] = self.name
        sort_order = ["name", "type", "host", "http_path", "username", "database", "schema"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}


class RedshiftConnection(BaseConnection):
    def __init__(
        self,
        name: str,
        host: str,
        username: str,
        password: str,
        port: int = 5439,
        database: str = None,
        schema: str = None,
        **kwargs,
    ) -> None:
        self.type = ConnectionType.redshift
        self.name = name
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database
        self.schema = schema

    def to_dict(self):
        base = {
            "name": self.name,
            "host": self.host,
            "port": self.port,
            "username": self.username,
            "password": self.password,
            "type": self.type,
        }
        if self.database:
            base["database"] = self.database
        if self.schema:
            base["schema"] = self.schema
        return base

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("password")
        attributes["name"] = self.name
        sort_order = ["name", "type", "host", "port", "username", "database", "schema"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}


class PostgresConnection(RedshiftConnection):
    def __init__(
        self,
        name: str,
        host: str,
        user: str,
        password: str,
        port: int = 5432,
        dbname: str = None,
        schema: str = None,
        **kwargs,
    ) -> None:
        self.type = ConnectionType.postgres
        self.name = name
        self.host = host
        self.port = port
        self.user = user
        self.username = user
        self.password = password
        self.database = dbname
        self.dbname = dbname
        self.schema = schema


class MySQLConnection(RedshiftConnection):
    def __init__(
        self,
        name: str,
        host: str,
        user: str,
        password: str,
        port: int = 3306,
        database: str = None,
        schema: str = None,
        **kwargs,
    ) -> None:
        self.type = ConnectionType.mysql
        self.name = name
        self.host = host
        self.port = port
        self.user = user
        self.username = user
        self.password = password
        self.database = database
        self.schema = schema


class DuckDBConnection(RedshiftConnection):
    def __init__(
        self,
        name: str,
        host: str,
        user: str,
        password: str,
        port: int = 5432,
        database: str = None,
        schema: str = None,
        **kwargs,
    ) -> None:
        self.type = ConnectionType.duck_db
        self.name = name
        self.host = host
        self.port = port
        self.user = user
        self.username = user
        self.password = password
        self.database = database
        self.schema = schema


class TrinoConnection(BaseConnection):
    def __init__(
        self,
        name: str,
        host: str,
        user: str,
        database: str,
        schema: str = None,
        scheme: str = "http",
        auth: Any = None,
        port: int = 8080,
        **kwargs,
    ) -> None:
        self.type = ConnectionType.trino
        self.name = name
        self.host = host
        self.user = user
        self.port = port
        self.auth = auth
        self.scheme = scheme
        self.database = database
        self.schema = schema

    def to_dict(self):
        base = {
            "name": self.name,
            "host": self.host,
            "port": self.port,
            "scheme": self.scheme,
            "type": self.type,
        }
        if self.user:
            base["user"] = self.user
        if self.auth:
            base["auth"] = self.auth
        return base

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("password")
        attributes["name"] = self.name
        sort_order = ["name", "type", "host", "port", "user", "database", "scheme"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}


class DruidConnection(BaseConnection):
    def __init__(
        self,
        name: str,
        host: str,
        username: str = None,
        user: str = None,
        password: str = None,
        port: int = 8082,
        path: str = "/druid/v2/sql/",
        scheme: str = "http",
        **kwargs,
    ) -> None:
        self.type = ConnectionType.druid
        self.name = name
        self.host = host
        self.port = port
        if user and username:
            raise ArgumentError(
                "Received arguments for both user and username, "
                "please send only one argument for the Druid user"
            )
        elif username:
            self.user = username
        elif user:
            self.user = user
        self.password = password
        self.path = path
        self.scheme = scheme
        self.database = None
        self.schema = None

    def to_dict(self):
        base = {
            "name": self.name,
            "host": self.host,
            "port": self.port,
            "path": self.path,
            "scheme": self.scheme,
            "type": self.type,
        }
        if self.user:
            base["user"] = self.user
        if self.password:
            base["password"] = self.password
        return base

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("password")
        attributes["name"] = self.name
        sort_order = ["name", "type", "host", "port", "user", "path", "scheme"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}


class SQLServerConnection(BaseConnection):
    def __init__(
        self,
        name: str,
        host: str,
        username: str = None,
        user: str = None,
        password: str = None,
        port: int = 1433,
        database: str = None,
        schema: str = None,
        **kwargs,
    ) -> None:
        self.type = ConnectionType.sql_server
        self.name = name
        self.host = host
        self.port = port
        if user and username:
            raise ArgumentError(
                "Received arguments for both user and username, "
                "please send only one argument for the SQL Server user"
            )
        elif username:
            self.user = username
        elif user:
            self.user = user
        self.password = password
        self.database = database
        self.schema = schema

    def to_dict(self):
        base = {
            "name": self.name,
            "host": self.host,
            "port": self.port,
            "type": self.type,
        }
        if self.user:
            base["user"] = self.user
        if self.password:
            base["password"] = self.password
        return base

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("password")
        attributes["name"] = self.name
        sort_order = ["name", "type", "host", "port", "user"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}


# The Synapse connection is the same as SQL Server
class AzureSynapseConnection(SQLServerConnection):
    pass


class BigQueryConnection(BaseConnection):
    def __init__(
        self, name: str, schema: str = None, credentials: str = None, keyfile: str = None, **kwargs
    ) -> None:
        self.type = ConnectionType.bigquery
        self.name = name
        self.schema = schema
        creds_to_use = credentials if credentials else keyfile
        self.credentials = self._convert_json_if_needed(creds_to_use, kwargs)
        self.project_id = self.credentials["project_id"]
        self.database = self.project_id

    def to_dict(self):
        """Dict for use with the BigQuery connector"""
        return {
            "name": self.name,
            "credentials": self.credentials,
            "project_id": self.project_id,
            "type": self.type,
        }

    def printable_attributes(self):
        attributes = deepcopy(self.to_dict())
        attributes.pop("credentials")
        attributes["name"] = self.name
        sort_order = ["name", "type", "project_id"]
        return {key: attributes.get(key) for key in sort_order if attributes.get(key) is not None}

    @staticmethod
    def _convert_json_if_needed(creds: dict, kwargs: dict):
        if isinstance(creds, dict):
            return deepcopy(creds)
        elif isinstance(creds, str):
            try:
                return json.loads(creds)
            except json.JSONDecodeError:
                # This means it's a file path not a JSON string
                if os.path.isabs(creds):
                    path = creds
                else:
                    path = os.path.join(kwargs["directory"], creds)

                with open(path, "r") as f:
                    return json.load(f)
        else:
            raise TypeError(f"BigQuery credentials json had wrong type: {type(creds)} for value {creds}")


connection_class_lookup = {
    ConnectionType.snowflake: SnowflakeConnection,
    ConnectionType.redshift: RedshiftConnection,
    ConnectionType.bigquery: BigQueryConnection,
    ConnectionType.postgres: PostgresConnection,
    ConnectionType.druid: DruidConnection,
    ConnectionType.sql_server: SQLServerConnection,
    ConnectionType.azure_synapse: AzureSynapseConnection,
    ConnectionType.duck_db: DuckDBConnection,
    ConnectionType.databricks: DatabricksConnection,
    ConnectionType.trino: TrinoConnection,
    ConnectionType.mysql: MySQLConnection,
}



import random
import string
import uuid


def generate_uuid(db_safe=False):
    if db_safe:
        return generate_random_password(40)
    return str(uuid.uuid4())


def generate_random_password(length):
    # Random string with the combination of lower and upper case
    letters = string.ascii_letters
    result_str = "".join(random.choice(letters) for i in range(length))
    return result_str


def flatten_filters(filters: list, return_nesting_depth: bool = False):
    nesting_depth = 0
    flat_list = []

    def recurse(filter_obj, return_nesting_depth: bool):
        nonlocal nesting_depth
        if isinstance(filter_obj, dict):
            if "conditions" in filter_obj:
                nesting_depth += 1
                for f in filter_obj["conditions"]:
                    recurse(f, return_nesting_depth)
            else:
                if return_nesting_depth:
                    filter_obj["nesting_depth"] = nesting_depth
                flat_list.append(filter_obj)
        elif isinstance(filter_obj, list):
            nesting_depth += 1
            for item in filter_obj:
                recurse(item, return_nesting_depth)

    recurse(filters, return_nesting_depth=return_nesting_depth)
    return flat_list



import difflib
import re
from typing import Iterable

from metrics_layer.core.exceptions import QueryError

NAME_REGEX = re.compile(r"([A-Za-z0-9\_]+)")


class MetricsLayerBase:
    def __init__(self, definition: dict = {}) -> None:
        self._definition = definition

    def __getattr__(self, attr: str):
        return self._definition.get(attr, None)

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}>"

    def to_dict(self):
        return {**self._definition}

    @staticmethod
    def valid_name(name: str):
        match = re.match(NAME_REGEX, name)
        if match is None:
            return False
        return match.group(1) == name

    @staticmethod
    def name_error(entity_name: str, name: str):
        return (
            f"{entity_name.title()} name: {name} is invalid. Please reference "
            "the naming conventions (only letters, numbers, or underscores)"
        )

    @staticmethod
    def _raise_query_error_from_cte(field_name: str):
        raise QueryError(
            f"Field {field_name} is not present in either source query, so it"
            " cannot be applied as a filter. Please add it to one of the source queries."
        )

    @staticmethod
    def line_col(element):
        line = getattr(getattr(element, "lc", None), "line", None)
        column = getattr(getattr(element, "lc", None), "col", None)
        return line, column

    @staticmethod
    def invalid_property_error(
        definition: dict, valid_properties: Iterable[str], entity_name: str, name: str, error_func: callable
    ):
        errors = []
        for key in definition:
            if key not in valid_properties:
                proposed_property = MetricsLayerBase.propose_property(key, valid_properties)
                proposed = f" Did you mean {proposed_property}?" if proposed_property else ""
                errors.append(
                    error_func(
                        definition[key],
                        (
                            f"Property {key} is present on {entity_name.title()} {name}, but it is not a"
                            f" valid property.{proposed}"
                        ),
                    )
                )
        return errors

    @staticmethod
    def field_name_parts(field_name: str):
        if field_name.count(".") == 1:
            view_name, name = field_name.split(".")
        else:
            view_name, name = None, field_name
        return view_name, name

    @staticmethod
    def propose_property(invalid_property_name: str, valid_properties: Iterable[str]) -> str:
        closest_match = difflib.get_close_matches(invalid_property_name, valid_properties, n=1)
        if closest_match:
            return closest_match[0]
        else:
            return ""


class SQLReplacement:
    @staticmethod
    def fields_to_replace(text: str):
        matches = re.finditer(r"\$\{(.*?)\}", text, re.MULTILINE)
        return [match.group(1) for match in matches]
